# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Custom types that can be used with click."""
from typing import Any, Optional

import click
from click import Context, Parameter

from pynessie.model import ContentKey


class ContentKeyParamType(click.ParamType):
    """ContentType click type."""

    name = "content_key"

    # pylint: disable=R1710
    def convert(self, value: Any, param: Optional[Parameter], ctx: Optional[Context]) -> ContentKey:
        """Convert incoming param into ContentKey type."""
        if isinstance(value, ContentKey):
            return value

        try:
            return ContentKey.from_path_string(str(value))
        except (TypeError, ValueError) as e:
            self.fail(f"{value!r} is not a valid content key due to: '{e!r}'", param, ctx)


CONTENT_KEY = ContentKeyParamType()



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Cli Common context functions that can be used by CLI commands/groups."""
from typing import Any, Dict, List, Mapping, Tuple

import attr
import click
from click import Option, UsageError

from pynessie.client import NessieClient


@attr.s(auto_attribs=True)
class ContextObject:
    """Click context object."""

    nessie: NessieClient
    verbose: bool
    json: bool


class MutuallyExclusiveOption(Option):
    """Only allow one option in a list to be set at once."""

    def __init__(self, *args: List, **kwargs: Dict) -> None:
        """Instantiated a mutually exclusive option."""
        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        super().__init__(*args, **kwargs)  # type: ignore

    def handle_parse_result(self, ctx: click.Context, opts: Mapping, args: List) -> Tuple[Any, List[str]]:
        """Ensure mutually exclusive options are not used together."""
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            raise UsageError(
                "Illegal usage: `{}` is mutually exclusive with " "arguments `{}`.".format(self.name, ", ".join(self.mutually_exclusive))
            )

        return super().handle_parse_result(ctx, opts, args)


class DefaultHelp(click.Command):
    """If no options are presented show help."""

    def __init__(self, *args: List, **kwargs: Dict) -> None:
        """Ensure that help is shown if nothing else is selected."""
        context_settings = kwargs.setdefault("context_settings", {})
        if "help_option_names" not in context_settings:
            context_settings["help_option_names"] = ["-h", "--help"]
        self.help_flag = context_settings["help_option_names"][0]
        super().__init__(*args, **kwargs)  # type: ignore

    def parse_args(self, ctx: click.Context, args: List) -> List:
        """Ensure that help is shown if nothing else is selected."""
        if not args:
            args = [self.help_flag]
        return super().parse_args(ctx, args)



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Contents List Command CLI."""

from collections import defaultdict
from typing import List

import click

from pynessie.cli_common_context import ContextObject, MutuallyExclusiveOption
from pynessie.decorators import error_handler, pass_client, validate_reference
from pynessie.model import Entries, Entry, EntrySchema
from pynessie.utils import build_filter_for_contents_listing_flags


@click.command("list")
@click.option("-r", "--ref", help="Branch to list from. If not supplied the default branch from config is used")
@click.option(
    "-t",
    "--type",
    "entity_types",
    help="entity types to filter on, if no entity types are passed then all types are returned",
    multiple=True,
)
@click.option(
    "--filter",
    "query_filter",
    multiple=False,
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["entity_type"],
    help="Allows advanced filtering using the Common Expression Language (CEL). "
    "An intro to CEL can be found at https://github.com/google/cel-spec/blob/master/doc/intro.md.\n"
    "Some examples with usable variables 'entry.namespace' (string) & 'entry.contentType' (string) are:\n"
    "entry.namespace.startsWith('a.b.c')\n"
    "entry.contentType in ['ICEBERG_TABLE','DELTA_LAKE_TABLE']\n"
    "entry.namespace.startsWith('some.name.space') && entry.contentType in ['ICEBERG_TABLE','DELTA_LAKE_TABLE']\n",
)
@pass_client
@error_handler
@validate_reference
def list_(ctx: ContextObject, ref: str, query_filter: str, entity_types: List[str]) -> None:
    """List content.

    Examples:

        nessie content list -r dev -> List all contents in 'dev' branch.

        nessie content list -r dev -t DELTA_LAKE_TABLE ->  List all contents in
    'dev' branch with type `DELTA_LAKE_TABLE`.

        nessie content list -r dev --filter "entry.namespace.startsWith('some.name.space')" -> List all contents in
    'dev' branch that start with 'some.name.space'
    """
    keys = ctx.nessie.list_keys(
        ref,
        query_filter=build_filter_for_contents_listing_flags(query_filter, entity_types),
    )
    results = EntrySchema().dumps(_format_keys_json(keys), many=True) if ctx.json else _format_keys(keys)
    click.echo(results)


def _format_keys_json(keys: Entries) -> List[Entry]:
    results = []
    for k in keys.entries:
        results.append(k)
    return results


def _format_keys(keys: Entries) -> str:
    results = defaultdict(list)
    result_str = ""
    for entry in keys.entries:
        results[entry.kind].append(entry.name)
    for k, result_list in results.items():
        result_str += k + ":\n"
        for v in result_list:
            value = ['"{}"'.format(i) if "." in i else i for i in v.elements]
            result_str += "\t{}\n".format(".".join(value))
    return result_str



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Tests for `config_parser.py`."""

import os

from assertpy import assert_that

from pynessie.conf.config_parser import build_config


def test_config_from_env() -> None:
    """Makes sure NESSIE environment variables are resolved automatically."""
    os.environ["NESSIE_TEST1_TEST2_TEST3"] = "test_val"
    config = build_config()
    assert_that(config["test1"]["test2"]["test3"].get()).is_equal_to("test_val")


def test_config_from_args() -> None:
    """Makes sure explicit config args are respected."""
    config = build_config({"abc.def": "test_val"})
    assert_that(config["abc"]["def"].get()).is_equal_to("test_val")


def test_args_take_precedence_over_env() -> None:
    """Makes sure explicit config args take precedence over environment variables."""
    os.environ["NESSIE_TEST1_TEST2_TEST3"] = "env_val"
    config = build_config({"test1.test2.test3": "arg_val"})
    assert_that(config["test1"]["test2"]["test3"].get()).is_equal_to("arg_val")



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""IO methods to read and write config."""
import os

import confuse


def write_to_file(config: confuse.Configuration) -> None:
    """Write updated config to file."""
    config_filename = os.path.join(config.config_dir(), confuse.CONFIG_FILENAME)
    with open(config_filename, "w", encoding="UTF-8") as f:
        f.write(config.dump())



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Contents View Command CLI."""
from typing import List

import click

from pynessie.cli_common_context import ContextObject
from pynessie.client import NessieClient
from pynessie.decorators import error_handler, pass_client, validate_reference
from pynessie.model import Content, ContentKey, ContentSchema
from pynessie.types import CONTENT_KEY


@click.command("view")
@click.option("-r", "--ref", help="Branch to list from. If not supplied the default branch from config is used.")
@click.argument("key", nargs=-1, required=True, type=CONTENT_KEY)
@pass_client
@error_handler
@validate_reference
def view(ctx: ContextObject, ref: str, key: List[ContentKey]) -> None:
    """View content.

        KEY is the content key that is associated with a specific content to view.
    This accepts as well multiple keys with space in between. The key can be in this format:
    'table.key' or 'namespace."table.key"'.

    Examples:

        nessie content view -r dev my_table -> View content details for content "my_table"
    in 'dev' branch.
    """
    if ctx.json:
        results = ContentSchema().dumps(_get_content_for_all_keys(ctx.nessie, ref, key), many=True)
    else:
        results = "\n".join((i.pretty_print() for i in _get_content_for_all_keys(ctx.nessie, ref, key)))

    click.echo(results)


def _get_content_for_all_keys(client: NessieClient, ref: str, keys: List[ContentKey]) -> List[Content]:
    contents: List[Content] = []

    for key in keys:
        contents.append(client.get_content(ref, key))

    return contents



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Top-level package for Nessie Contents CLI commands."""

import click

from pynessie.cli_common_context import ContextObject
from pynessie.commands.content.commit import commit
from pynessie.commands.content.list_ import list_
from pynessie.commands.content.view import view
from pynessie.decorators import pass_client


@click.group(name="content")
@pass_client
def content(ctx: ContextObject) -> None:
    """View, list content, and commit changes."""
    pass


content.add_command(list_)
content.add_command(view)
content.add_command(commit)



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Parser for confuse Configuration object."""
import os
from typing import Optional

import confuse


def _get_env_args() -> dict:
    args = {}
    for k, v in os.environ.items():
        if "NESSIE_" in k and k != "NESSIE_CLIENTDIR":
            name = k.replace("NESSIE_", "").lower().replace("_", ".")
            if name == "auth.timeout":
                v = int(v)  # type: ignore
            args[name] = v
    return args


def build_config(args: Optional[dict] = None) -> confuse.Configuration:
    """Build configuration object from input params, env variables and yaml file."""
    config = confuse.Configuration("nessie", __name__)
    env_args = _get_env_args()
    config.set_args(env_args, dots=True)
    if args:
        config.set_args(args, dots=True)
    config["auth"]["password"].redact = True
    return config



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Makes sure the expression building functions work properly."""

from typing import List

from assertpy import assert_that

from pynessie.utils import (
    build_filter_for_commit_log_flags,
    build_filter_for_contents_listing_flags,
    parse_to_iso8601,
)


def test_building_empty_filter() -> None:
    """Makes sure the expression building functions work with empty input."""
    assert_that(build_filter_for_commit_log_flags(query_filter="", authors=[], committers=[], since=None, until=None)).is_none()

    assert_that(build_filter_for_contents_listing_flags(query_filter="", types=[])).is_none()


def test_building_filter_for_commit_log() -> None:
    """Makes sure the filter building function for the commit log produces what we expect."""
    authors: List[str] = []
    committers: List[str] = []
    since = None
    until = None
    query_expr = ""
    assert_that(
        build_filter_for_commit_log_flags(
            query_filter="some_exclusive_filter", authors=authors, committers=committers, since=since, until=until
        )
    ).is_equal_to("some_exclusive_filter")

    assert_that(
        build_filter_for_commit_log_flags(
            query_filter="some_exclusive_filter", authors=["one"], committers=["two"], since=since, until=until
        )
    ).is_equal_to("some_exclusive_filter")

    assert_that(
        build_filter_for_commit_log_flags(query_filter=query_expr, authors=["one"], committers=committers, since=since, until=until)
    ).is_equal_to("commit.author=='one'")

    assert_that(
        build_filter_for_commit_log_flags(
            query_filter=query_expr, authors=["one", "two", "three"], committers=committers, since=since, until=until
        )
    ).is_equal_to("(commit.author=='one' || commit.author=='two' || commit.author=='three')")

    assert_that(
        build_filter_for_commit_log_flags(query_filter=query_expr, authors=["one"], committers=["two"], since=since, until=until)
    ).is_equal_to("(commit.author=='one' && commit.committer=='two')")

    assert_that(
        build_filter_for_commit_log_flags(
            query_filter=query_expr, authors=["one", "two"], committers=["two", "three"], since=since, until=until
        )
    ).is_equal_to("((commit.author=='one' || commit.author=='two') && (commit.committer=='two' || commit.committer=='three'))")

    assert_that(
        build_filter_for_commit_log_flags(query_filter=query_expr, authors=[], committers=[], since="2021-05-31T08:23:15Z", until=until)
    ).is_equal_to("timestamp(commit.commitTime) > timestamp('2021-05-31T08:23:15+00:00')")

    assert_that(
        build_filter_for_commit_log_flags(query_filter=query_expr, authors=[], committers=[], since="", until="2021-05-31T10:23:15Z")
    ).is_equal_to("timestamp(commit.commitTime) < timestamp('2021-05-31T10:23:15+00:00')")

    assert_that(
        build_filter_for_commit_log_flags(
            query_filter=query_expr, authors=[], committers=[], since="2021-05-31T08:23:15Z", until="2021-05-31T10:23:15Z"
        )
    ).is_equal_to(
        "(timestamp(commit.commitTime) > timestamp('2021-05-31T08:23:15+00:00') "
        "&& timestamp(commit.commitTime) < timestamp('2021-05-31T10:23:15+00:00'))"
    )

    assert_that(
        build_filter_for_commit_log_flags(
            query_filter=query_expr,
            authors=["one", "two"],
            committers=["two", "three"],
            since="2021-05-31T08:23:15Z",
            until="2021-05-31T10:23:15Z",
        )
    ).is_equal_to(
        "((commit.author=='one' || commit.author=='two') "
        "&& (commit.committer=='two' || commit.committer=='three') "
        "&& timestamp(commit.commitTime) > timestamp('2021-05-31T08:23:15+00:00') "
        "&& timestamp(commit.commitTime) < timestamp('2021-05-31T10:23:15+00:00'))"
    )


def test_building_filter_for_content_entries() -> None:
    """Makes sure the expression building function for the content entries produces what we expect."""
    assert_that(build_filter_for_contents_listing_flags(query_filter="exclusive_expr", types=[])).is_equal_to("exclusive_expr")

    assert_that(
        build_filter_for_contents_listing_flags(query_filter="exclusive_expr", types=["ICEBERG_VIEW", "DELTA_LAKE_TABLE"])
    ).is_equal_to("exclusive_expr")

    assert_that(build_filter_for_contents_listing_flags(query_filter="", types=["ICEBERG_TABLE"])).is_equal_to(
        "entry.contentType in ['ICEBERG_TABLE']"
    )
    assert_that(build_filter_for_contents_listing_flags(query_filter="", types=["ICEBERG_TABLE", "DELTA_LAKE_TABLE"])).is_equal_to(
        "entry.contentType in ['ICEBERG_TABLE','DELTA_LAKE_TABLE']"
    )


def test_date_parsing() -> None:
    """Tests date parsing."""
    assert_that(parse_to_iso8601("2021-05-31 08:23:15Z")).is_equal_to("2021-05-31T08:23:15+00:00")
    assert_that(parse_to_iso8601("2021-05-31T08:23:15+01:00")).is_equal_to("2021-05-31T07:23:15+00:00")
    assert_that(parse_to_iso8601("2021-05-31T08:23:15+02:00")).is_equal_to("2021-05-31T06:23:15+00:00")
    assert_that(parse_to_iso8601("2021-05-31 08:23:15Z+00:00")).is_equal_to("2021-05-31T08:23:15+00:00")
    assert_that(parse_to_iso8601("2021-06-30T08:19:04.826051Z")).is_equal_to("2021-06-30T08:19:04.826051+00:00")


def test_invalid_date_parsing() -> None:
    """Tests invalid date parsing."""
    assert_that(parse_to_iso8601).raises(ValueError).when_called_with(None)
    assert_that(parse_to_iso8601).raises(ValueError).when_called_with("")
    assert_that(parse_to_iso8601).raises(ValueError).when_called_with("some_invalid_str")
    assert_that(parse_to_iso8601).raises(ValueError).when_called_with("99999")



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Console script for nessie_client."""
import sys
from typing import Any

import click
import confuse

from pynessie import __version__
from pynessie.cli_common_context import ContextObject
from pynessie.client import NessieClient
from pynessie.commands import (
    branch_,
    cherry_pick,
    config,
    content,
    diff,
    log,
    merge,
    reflog,
    remote,
    tag,
)
from pynessie.conf import build_config


def _print_version(ctx: Any, param: Any, value: Any) -> None:
    if not value or ctx.resilient_parsing:
        return
    click.echo("nessie version " + __version__)
    ctx.exit()


@click.group("nessie")
@click.option("--json", is_flag=True, help="write output in json format.")
@click.option("-v", "--verbose", is_flag=True, help="Verbose output.")
@click.option("--endpoint", help="Optional endpoint, if different from config file.")
@click.option("--auth-token", help="Optional bearer auth token, if different from config file.")
@click.option("--version", is_flag=True, callback=_print_version, expose_value=False, is_eager=True)
@click.pass_context
def cli(ctx: click.core.Context, json: bool, verbose: bool, endpoint: str, auth_token: str) -> None:
    """Nessie cli tool.

    Interact with Nessie branches and tables via the command line
    """
    try:
        cfg_map = {}
        if endpoint:
            cfg_map["endpoint"] = endpoint
        if auth_token:
            cfg_map["auth.type"] = "bearer"
            cfg_map["auth.token"] = auth_token
        nessie = NessieClient(build_config(cfg_map))
        ctx.obj = ContextObject(nessie, verbose, json)
    except confuse.exceptions.ConfigTypeError as e:
        raise click.ClickException(str(e)) from e


cli.add_command(remote)
cli.add_command(config)
cli.add_command(log)
cli.add_command(branch_)
cli.add_command(tag)
cli.add_command(merge)
cli.add_command(cherry_pick)
cli.add_command(content)
cli.add_command(diff)
cli.add_command(reflog)


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    sys.exit(cli())  # pragma: no cover



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Use AWS4Auth and botocore to fetch credentials and sign requests."""
from typing import Optional

from botocore.credentials import get_credentials
from botocore.session import Session
from requests_aws4auth import AWS4Auth


def setup_aws_auth(region: str, profile: Optional[str] = None) -> AWS4Auth:
    """For a given region sign a request to the execute-api with standard credentials."""
    credentials = get_credentials(Session(profile=profile))
    auth = AWS4Auth(credentials.access_key, credentials.secret_key, region, "execute-api", session_token=credentials.token)
    return auth



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for `content CLI command`."""
from typing import List, Optional

import pytest
import simplejson
from assertpy import assert_that

from pynessie.model import (
    ContentSchema,
    DeltaLakeTable,
    EntrySchema,
    IcebergTable,
    IcebergView,
    ReferenceSchema,
)

from .conftest import execute_cli_command, make_commit, ref_hash

CONTENT_COMMAND = "content"


@pytest.mark.vcr
def test_content_view() -> None:
    """Test content view."""
    branch = "contents_view_dev"
    execute_cli_command(["branch", branch])

    iceberg_table = _create_iceberg_table("test_contents_view")
    delta_lake_table = _create_delta_lake_table("test_dl_table")
    iceberg_view = _create_iceberg_view("test_iceberg_view")

    make_commit("this.is.iceberg.foo", iceberg_table, branch)
    make_commit("this.is.delta.bar", delta_lake_table, branch)
    make_commit('this.is."sql.baz"', iceberg_view, branch)

    result_table = ContentSchema().loads(
        execute_cli_command(["--json", CONTENT_COMMAND, "view", "--ref", branch, "this.is.iceberg.foo"]), many=True
    )

    assert_that(result_table).is_length(1)
    assert_that(result_table[0]).is_equal_to(iceberg_table)

    branch_hash = ref_hash(branch)
    result_table_2 = ContentSchema().loads(
        execute_cli_command(["--json", CONTENT_COMMAND, "view", "--ref", branch_hash, "this.is.iceberg.foo"]), many=True
    )
    assert_that(result_table_2).is_equal_to(result_table)
    result_table_2 = ContentSchema().loads(
        execute_cli_command(["--json", CONTENT_COMMAND, "view", "--ref", f"{branch}@{branch_hash}", "this.is.iceberg.foo"]), many=True
    )
    assert_that(result_table_2).is_equal_to(result_table)

    result_table = ContentSchema().loads(
        execute_cli_command(["--json", CONTENT_COMMAND, "view", "--ref", branch, "this.is.delta.bar"]), many=True
    )

    assert_that(result_table).is_length(1)
    assert_that(result_table[0]).is_equal_to(delta_lake_table)

    result_table = ContentSchema().loads(
        execute_cli_command(["--json", CONTENT_COMMAND, "view", "--ref", branch, 'this.is."sql.baz"']), many=True
    )

    assert_that(result_table).is_length(1)
    assert_that(result_table[0]).is_equal_to(iceberg_view)


@pytest.mark.vcr
def test_content_list() -> None:
    """Test content list."""
    branch = "contents_list_dev"
    execute_cli_command(["branch", branch])

    iceberg_table = _create_iceberg_table("test_contents_list")
    delta_lake_table = _create_delta_lake_table("test_dl_table_list")
    iceberg_view = _create_iceberg_view("test_iceberg_view_list")

    make_commit("this.is.iceberg.foo", iceberg_table, branch)
    make_commit("this.is.delta.bar", delta_lake_table, branch)
    make_commit("this.is.sql.baz", iceberg_view, branch)

    tables = EntrySchema().loads(
        execute_cli_command(["--json", CONTENT_COMMAND, "list", "--ref", branch, "--type", "ICEBERG_TABLE"]), many=True
    )
    assert_that(tables).is_length(1)
    assert_that(tables[0].kind).is_equal_to("ICEBERG_TABLE")

    branch_hash = ref_hash(branch)
    tables_2 = EntrySchema().loads(
        execute_cli_command(["--json", CONTENT_COMMAND, "list", "--ref", branch_hash, "--type", "ICEBERG_TABLE"]), many=True
    )
    assert_that(tables_2).is_equal_to(tables)
    tables_2 = EntrySchema().loads(
        execute_cli_command(["--json", CONTENT_COMMAND, "list", "--ref", f"{branch}@{branch_hash}", "--type", "ICEBERG_TABLE"]), many=True
    )
    assert_that(tables_2).is_equal_to(tables)

    tables = EntrySchema().loads(
        execute_cli_command(["--json", CONTENT_COMMAND, "list", "--ref", branch, "--type", "DELTA_LAKE_TABLE"]), many=True
    )
    assert_that(tables).is_length(1)
    assert_that(tables[0].kind).is_equal_to("DELTA_LAKE_TABLE")

    result = execute_cli_command(["--json", CONTENT_COMMAND, "list", "--ref", branch, "--filter", "entry.contentType == 'ICEBERG_TABLE'"])
    tables = EntrySchema().loads(result, many=True)
    assert_that(tables).is_length(1)
    assert_that(tables[0].kind).is_equal_to("ICEBERG_TABLE")

    result = execute_cli_command(
        ["--json", CONTENT_COMMAND, "list", "--ref", branch, "--filter", "entry.contentType in ['ICEBERG_TABLE', 'DELTA_LAKE_TABLE']"]
    )
    tables = EntrySchema().loads(result, many=True)
    assert_that(tables).is_length(2)
    assert_that(set(t.kind for t in tables)).is_equal_to({"DELTA_LAKE_TABLE", "ICEBERG_TABLE"})

    result = execute_cli_command(
        ["--json", CONTENT_COMMAND, "list", "--ref", branch, "--filter", "entry.namespace.startsWith('this.is.del')"]
    )
    tables = EntrySchema().loads(result, many=True)
    assert_that(tables).is_length(1)
    assert_that(tables[0].kind).is_equal_to("DELTA_LAKE_TABLE")

    result = execute_cli_command(["--json", CONTENT_COMMAND, "list", "--ref", branch, "--filter", "entry.namespace.startsWith('this.is')"])
    tables = EntrySchema().loads(result, many=True)
    assert_that(tables).is_length(3)
    assert_that(set(i.kind for i in tables)).is_equal_to({"ICEBERG_TABLE", "ICEBERG_VIEW", "DELTA_LAKE_TABLE"})


@pytest.mark.vcr
def test_content_commit_delete() -> None:
    """Test content commit delete operation."""
    branch = "contents_commit_delete_dev"
    execute_cli_command(["branch", branch])

    iceberg_table = _create_iceberg_table("test_contents_delete")

    make_commit("this.is.iceberg.foo", iceberg_table, branch)

    tables = EntrySchema().loads(
        execute_cli_command(["--json", CONTENT_COMMAND, "list", "--ref", branch, "--type", "ICEBERG_TABLE"]), many=True
    )
    assert_that(tables).is_length(1)

    head_hash = _get_head_branch_hash(branch)

    execute_cli_command(
        [CONTENT_COMMAND, "commit", "-R", "--ref", branch, "-m", "delete test table", "-c", head_hash, "this.is.iceberg.foo"]
    )

    tables = EntrySchema().loads(
        execute_cli_command(["--json", CONTENT_COMMAND, "list", "--ref", branch, "--type", "ICEBERG_TABLE"]), many=True
    )
    assert_that(tables).is_length(0)


@pytest.mark.vcr
def test_content_commit_with_empty_content_delete() -> None:
    """Test content commit delete with empty content operation."""
    branch = "contents_commit_with_empty_content_delete_dev"
    execute_cli_command(["branch", branch])

    iceberg_table = _create_iceberg_table("test_contents_with_empty_content_delete")

    make_commit("this.is.iceberg.foo", iceberg_table, branch)

    tables = EntrySchema().loads(
        execute_cli_command(["--json", CONTENT_COMMAND, "list", "--ref", branch, "--type", "ICEBERG_TABLE"]), many=True
    )
    assert_that(tables).is_length(1)

    head_hash = _get_head_branch_hash(branch)

    execute_cli_command(
        [CONTENT_COMMAND, "commit", "--stdin", "this.is.iceberg.foo", "--ref", branch, "-m", "delete table", "-c", head_hash],
        input_data="",
    )

    tables = EntrySchema().loads(
        execute_cli_command(["--json", CONTENT_COMMAND, "list", "--ref", branch, "--type", "ICEBERG_TABLE"]), many=True
    )
    assert_that(tables).is_length(0)


@pytest.mark.vcr
def test_content_commit_with_edited_data() -> None:
    """Test content commit to edit data operation."""
    branch = "content_commit_with_edited_data_dev"
    execute_cli_command(["branch", branch])

    iceberg_table = _create_iceberg_table("test_content_commit_with_edited_data")
    table_id = "this.is.iceberg.foo"

    make_commit(table_id, iceberg_table, branch)

    edited_iceberg_table = _create_iceberg_table("test_content_commit_with_edited_data", metadata_location="/d/e/f")

    tables = EntrySchema().loads(
        execute_cli_command(["--json", CONTENT_COMMAND, "list", "--ref", branch, "--type", "ICEBERG_TABLE"]), many=True
    )
    assert_that(tables).is_length(1)

    # If we have the same content data, we expect no commit to happen
    make_commit(table_id, iceberg_table, branch)
    logs = simplejson.loads(execute_cli_command(["--json", "log", branch]))

    assert len(logs) == 1

    # Now we commit a modified table
    make_commit(table_id, edited_iceberg_table, branch)

    tables = EntrySchema().loads(
        execute_cli_command(["--json", CONTENT_COMMAND, "list", "--ref", branch, "--type", "ICEBERG_TABLE"]), many=True
    )
    assert_that(tables).is_length(1)

    result_table = ContentSchema().loads(execute_cli_command(["--json", CONTENT_COMMAND, "view", "--ref", branch, table_id]), many=True)

    assert_that(result_table).is_length(1)
    assert_that(result_table[0]).is_equal_to(edited_iceberg_table)


@pytest.mark.vcr
def test_content_commit_with_expected_state() -> None:
    """Test making a commit with some expected Contents, i.e. IcebergTable."""
    branch = "contents_commit_with_expected_state"
    execute_cli_command(["branch", branch])

    make_commit("commit.expected.contents", _create_iceberg_table("test_expected_contents"), branch, message="commit 1")
    # the second commit will use the Contents of the first one as "expected contents"
    make_commit("commit.expected.contents", _create_iceberg_table("test_expected_contents"), branch, message="commit 2")


@pytest.mark.vcr
def test_content_commit_no_expected_state() -> None:
    """Test making two commit without any expected Contents, i.e. using DeltaLakeTable."""
    branch = "contents_commit_with_no__expected_state"
    execute_cli_command(["branch", branch])

    table1 = _create_delta_lake_table("test_commit_no_expected_state", metadata_location_history=["asd111"])
    make_commit("commit.expected.contents", table1, branch, message="commit 1")

    table2 = _create_delta_lake_table("test_commit_no_expected_state", metadata_location_history=["asd222"])

    # the second commit will set new contents without the "expected contents" parameter due to using a DeltaLakeTable
    make_commit("commit.expected.contents", table2, branch, message="commit 2")


def _create_iceberg_table(
    table_id: str, metadata_location: str = "/a/b/c", snapshot_id: int = 42, schema_id: int = 42, spec_id: int = 42, sort_order_id: int = 42
) -> IcebergTable:
    return IcebergTable(table_id, metadata_location, snapshot_id, schema_id, spec_id, sort_order_id)


def _create_delta_lake_table(
    table_id: str,
    metadata_location_history: Optional[List[str]] = None,
    checkpoint_location_history: Optional[List[str]] = None,
    last_checkpoint: str = "x",
) -> DeltaLakeTable:
    if checkpoint_location_history is None:
        checkpoint_location_history = ["def"]
    if metadata_location_history is None:
        metadata_location_history = ["asd"]
    return DeltaLakeTable(table_id, last_checkpoint, checkpoint_location_history, metadata_location_history)


def _create_iceberg_view(
    table_id: str,
    metadata_location: str = "/a/b/c",
    version_id: int = 1,
    schema_id: int = 1,
    sql_text: str = "SELECT * FROM foo",
    dialect: str = "SPARK",
) -> IcebergView:
    return IcebergView(table_id, metadata_location, version_id, schema_id, dialect, sql_text)


def _get_head_branch_hash(branch: str) -> str:
    refs = {i.name: i.hash_ for i in ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)}
    return refs[branch]



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Nessie Data objects."""
import re
from datetime import datetime
from typing import List, Optional, Tuple

import attr
import desert
from marshmallow import fields
from marshmallow_oneofschema import OneOfSchema

# regex taken from org.projectnessie.model.Validation
__RE_REFERENCE_NAME_RAW = "[A-Za-z](((?![.][.])[A-Za-z0-9./_-])*[A-Za-z0-9._-])"
__RE_REFERENCE_NAME: re.Pattern = re.compile(f"^{__RE_REFERENCE_NAME_RAW}$")
__RE_HASH_RAW = "[0-9a-fA-F]{8,64}"
__RE_HASH: re.Pattern = re.compile(f"^{__RE_HASH_RAW}$")
__RE_REFERENCE_WITH_HASH: re.Pattern = re.compile(f"^({__RE_REFERENCE_NAME_RAW})([@]|[*])({__RE_HASH_RAW})$")


DETACHED_REFERENCE_NAME = "DETACHED"


def is_valid_reference_name(ref: str) -> bool:
    """Checks whether 'ref' is a valid reference name."""
    return __RE_REFERENCE_NAME.match(ref) is not None


def is_valid_hash(ref: str) -> bool:
    """Checks whether 'ref' is a valid commit id/hash."""
    return __RE_HASH.match(ref) is not None


def split_into_reference_and_hash(ref_with_hash: Optional[str]) -> Tuple[str, Optional[str]]:
    """Returns a tuple of reference-name + hash, if the given string represents a ref-name + hash tuple 'ref_name@commit_id'."""
    if not ref_with_hash:
        return "<UNKNOWN>", None
    match = __RE_REFERENCE_WITH_HASH.match(ref_with_hash)
    if not match:
        if is_valid_hash(ref_with_hash):
            return DETACHED_REFERENCE_NAME, ref_with_hash
        return ref_with_hash, None
    return match.group(1), match.group(5)


@attr.dataclass
class Content:
    """Dataclass for Nessie Content."""

    id: str = desert.ib(fields.Str())

    @staticmethod
    def requires_expected_state() -> bool:
        """Checks whether this Content object requires the "expected" state to be provided for Put operations."""
        return False

    def pretty_print(self) -> str:
        """Print out for cli."""
        raise NotImplementedError


@attr.dataclass
class IcebergTable(Content):
    """Dataclass for Nessie Content."""

    metadata_location: str = desert.ib(fields.Str(data_key="metadataLocation"))
    snapshot_id: int = desert.ib(fields.Int(data_key="snapshotId"))
    schema_id: int = desert.ib(fields.Int(data_key="schemaId"))
    spec_id: int = desert.ib(fields.Int(data_key="specId"))
    sort_order_id: int = desert.ib(fields.Int(data_key="sortOrderId"))

    @staticmethod
    def requires_expected_state() -> bool:
        """Returns True - expected state should be provided for Put operations on Iceberg tables."""
        return True

    def pretty_print(self) -> str:
        """Print out for cli."""
        return (
            f"Iceberg table:\n\tmetadata-location: {self.metadata_location}\n\tsnapshot-id: {self.snapshot_id}"
            f"\n\tschema-id: {self.schema_id}"
            f"\n\tpartition-spec-id: {self.spec_id}\n\tdefault-sort-order-id: {self.sort_order_id}"
        )


IcebergTableSchema = desert.schema_class(IcebergTable)


@attr.dataclass
class DeltaLakeTable(Content):
    """Dataclass for Nessie Content."""

    last_checkpoint: str = desert.ib(fields.Str(data_key="lastCheckpoint"))
    checkpoint_location_history: List[str] = desert.ib(fields.List(fields.Str(), data_key="checkpointLocationHistory"))
    metadata_location_history: List[str] = desert.ib(fields.List(fields.Str(), data_key="metadataLocationHistory"))

    def pretty_print(self) -> str:
        """Print out for cli."""
        deltas = "\n\t\t".join(self.metadata_location_history)
        checkpoints = "\n\t\t".join(self.checkpoint_location_history)
        return "DeltaLake table:\n\tLast Checkpoint: {}\n\tDelta History: {}\n\tCheckpoint History: {}".format(
            self.last_checkpoint, deltas, checkpoints
        )


DeltaLakeTableSchema = desert.schema_class(DeltaLakeTable)


@attr.dataclass
class IcebergView(Content):
    """Dataclass for Nessie Iceberg View."""

    metadata_location: str = desert.ib(fields.Str(data_key="metadataLocation"))
    version_id: int = desert.ib(fields.Int(data_key="versionId"))
    schema_id: int = desert.ib(fields.Int(data_key="schemaId"))
    dialect: str = desert.ib(fields.Str())
    sql_text: str = desert.ib(fields.Str(data_key="sqlText"))

    def pretty_print(self) -> str:
        """Print out for cli."""
        return "IcebergView:\n\tmetadata-location: {}\n\tversion-id: {}\n\tschema-id: {}\n\tDialect: {}\n\tSql: {}".format(
            self.metadata_location, self.version_id, self.schema_id, self.dialect, self.sql_text
        )  # todo use a sql parser to pretty print this


IcebergViewSchema = desert.schema_class(IcebergView)


class ContentSchema(OneOfSchema):
    """Schema for Nessie Content."""

    type_schemas = {
        "ICEBERG_TABLE": IcebergTableSchema,
        "DELTA_LAKE_TABLE": DeltaLakeTableSchema,
        "ICEBERG_VIEW": IcebergViewSchema,
    }

    def get_obj_type(self, obj: Content) -> str:
        """Returns the object type based on its class."""
        if isinstance(obj, IcebergTable):
            return "ICEBERG_TABLE"
        if isinstance(obj, DeltaLakeTable):
            return "DELTA_LAKE_TABLE"
        if isinstance(obj, IcebergView):
            return "ICEBERG_VIEW"

        raise ValueError("Unknown object type: {}".format(obj.__class__.__name__))


@attr.dataclass
class ContentKey:
    """ContentKey."""

    elements: List[str] = desert.ib(fields.List(fields.Str))

    def to_string(self) -> str:
        """Convert this key to friendly CLI string."""
        # false positives in pylint
        # pylint: disable=E1133
        return ".".join(f'"{i}"' if "." in i else i for i in self.elements)  # noqa: B028,B907

    def to_path_string(self) -> str:
        """Convert this key to a url encoded path string."""
        # false positives in pylint
        # pylint: disable=E1133
        return ".".join(i.replace(".", "\00") if i else "" for i in self.elements)

    @staticmethod
    def from_path_string(key: str) -> "ContentKey":
        """Convert from path encoded string to normal string."""
        return ContentKey([i for i in ContentKey._split_key_based_on_regex(key) if i])

    @staticmethod
    def _split_key_based_on_regex(raw_key: str) -> List[str]:
        # Find all occurrences of strings between double quotes
        # E.g: a.b."c.d"
        regex = re.compile('"[^"]*"')

        # Replace any dot that is inside double quotes with null char '\00' and remove the double quotes
        key_with_null = regex.sub(lambda x: x.group(0).replace(".", "\00").replace('"', ""), raw_key)

        # Split based on the dot
        splitted_key = key_with_null.split(".")

        # Return back the splitted elements and make sure to change back '/0' to '.'
        return [i.replace("\00", ".") for i in splitted_key]


ContentKeySchema = desert.schema_class(ContentKey)


@attr.dataclass
class Operation:
    """Single Commit Operation."""

    key: ContentKey = desert.ib(fields.Nested(ContentKeySchema))

    def pretty_print(self) -> str:
        """Print out for cli."""
        raise NotImplementedError


@attr.dataclass
class Put(Operation):
    """Single Commit Operation."""

    content: Content = desert.ib(fields.Nested(ContentSchema))
    expectedContent: Optional[Content] = attr.ib(default=None, metadata=desert.metadata(fields.Nested(ContentSchema, allow_none=True)))

    def pretty_print(self) -> str:
        """Print out for cli."""
        # pylint: disable=E1101
        return f"Put of {self.key.to_string()} : {self.content.pretty_print()}"


PutOperationSchema = desert.schema_class(Put)


@attr.dataclass
class Delete(Operation):
    """Delete single key."""

    def pretty_print(self) -> str:
        """Print out for cli."""
        # pylint: disable=E1101
        return f"Delete of {self.key.to_string()}"


DeleteOperationSchema = desert.schema_class(Delete)


@attr.dataclass
class Unchanged(Operation):
    """Unchanged single key."""

    def pretty_print(self) -> str:
        """Print out for cli."""
        # pylint: disable=E1101
        return f"Unchanged of {self.key.to_string()}"


UnchangedOperationSchema = desert.schema_class(Unchanged)


class OperationsSchema(OneOfSchema):
    """Schema for Nessie Operations."""

    type_schemas = {
        "PUT": PutOperationSchema,
        "UNCHANGED": UnchangedOperationSchema,
        "DELETE": DeleteOperationSchema,
    }

    def get_obj_type(self, obj: Operation) -> str:
        """Returns the object type based on its class."""
        if isinstance(obj, Put):
            return "PUT"
        if isinstance(obj, Unchanged):
            return "UNCHANGED"
        if isinstance(obj, Delete):
            return "DELETE"

        raise ValueError("Unknown object type: {}".format(obj.__class__.__name__))


@attr.dataclass
class CommitMeta:
    """Dataclass for commit metadata."""

    hash_: str = attr.ib(default=None, metadata=desert.metadata(fields.Str(data_key="hash")))
    commitTime: datetime = attr.ib(default=None, metadata=desert.metadata(fields.DateTime()))
    authorTime: datetime = attr.ib(default=None, metadata=desert.metadata(fields.DateTime()))
    committer: str = attr.ib(default=None, metadata=desert.metadata(fields.Str(allow_none=True)))
    author: str = attr.ib(default=None, metadata=desert.metadata(fields.Str(allow_none=True)))
    signedOffBy: str = attr.ib(default=None, metadata=desert.metadata(fields.Str(allow_none=True)))
    message: str = attr.ib(default=None, metadata=desert.metadata(fields.Str(allow_none=True)))
    properties: dict = desert.ib(fields.Dict(), default=None)


CommitMetaSchema = desert.schema_class(CommitMeta)


@attr.dataclass
class ReferenceMetadata:
    """Dataclass for Nessie ReferenceMetadata."""

    commit_meta_of_head: CommitMeta = desert.ib(fields.Nested(CommitMetaSchema, data_key="commitMetaOfHEAD", allow_none=True))
    num_commits_ahead: int = attr.ib(default=None, metadata=desert.metadata(fields.Int(allow_none=True, data_key="numCommitsAhead")))
    num_commits_behind: int = attr.ib(default=None, metadata=desert.metadata(fields.Int(allow_none=True, data_key="numCommitsBehind")))
    common_ancestor_hash: str = attr.ib(default=None, metadata=desert.metadata(fields.Str(allow_none=True, data_key="commonAncestorHash")))
    num_total_commits: str = attr.ib(default=None, metadata=desert.metadata(fields.Int(allow_none=True, data_key="numTotalCommits")))


ReferenceMetadataSchema = desert.schema_class(ReferenceMetadata)


@attr.dataclass
class Reference:
    """Dataclass for Nessie Reference."""

    name: str = desert.ib(fields.Str())
    hash_: Optional[str] = attr.ib(default=None, metadata=desert.metadata(fields.Str(data_key="hash", allow_none=True)))
    metadata: Optional[ReferenceMetadata] = attr.ib(
        default=None, metadata=desert.metadata(fields.Nested(ReferenceMetadataSchema, allow_none=True, data_key="metadata"))
    )


@attr.dataclass
class Detached(Reference):
    """Dataclass for Nessie detached commit id."""

    pass


DetachedSchema = desert.schema_class(Detached)


@attr.dataclass
class Branch(Reference):
    """Dataclass for Nessie Branch."""

    pass


BranchSchema = desert.schema_class(Branch)


@attr.dataclass
class Tag(Reference):
    """Dataclass for Nessie Tag."""

    pass


TagSchema = desert.schema_class(Tag)


class ReferenceSchema(OneOfSchema):
    """Schema for Nessie Reference."""

    type_schemas = {
        "BRANCH": BranchSchema,
        "TAG": TagSchema,
        "DETACHED": DetachedSchema,
    }

    def get_obj_type(self, obj: Reference) -> str:
        """Returns the object type based on its class."""
        if isinstance(obj, Branch):
            return "BRANCH"
        if isinstance(obj, Tag):
            return "TAG"
        if isinstance(obj, Detached):
            return "DETACHED"

        raise ValueError("Unknown object type: {}".format(obj.__class__.__name__))


@attr.dataclass
class ReferencesResponse:
    """Dataclass for References."""

    references: List[Reference] = desert.ib(fields.List(fields.Nested(ReferenceSchema())))
    has_more: bool = attr.ib(default=False, metadata=desert.metadata(fields.Bool(allow_none=True, data_key="hasMore")))
    token: str = attr.ib(default=None, metadata=desert.metadata(fields.Str(allow_none=True)))


ReferencesResponseSchema = desert.schema_class(ReferencesResponse)


@attr.dataclass
class EntryName:
    """Dataclass for Nessie Entry Name."""

    elements: List[str] = desert.ib(fields.List(fields.Str()))


EntryNameSchema = desert.schema_class(EntryName)


@attr.dataclass
class Entry:
    """Dataclass for Nessie Entry."""

    kind: str = desert.ib(fields.Str(data_key="type"))
    name: EntryName = desert.ib(fields.Nested(EntryNameSchema))


EntrySchema = desert.schema_class(Entry)


@attr.dataclass
class Entries:
    """Dataclass for Content Entries."""

    entries: List[Entry] = desert.ib(fields.List(fields.Nested(EntrySchema())))
    has_more: bool = attr.ib(default=False, metadata=desert.metadata(fields.Bool(allow_none=True, data_key="hasMore")))
    token: str = attr.ib(default=None, metadata=desert.metadata(fields.Str(allow_none=True)))


EntriesSchema = desert.schema_class(Entries)


@attr.dataclass
class LogEntry:
    """Dataclass for commit log entries."""

    commit_meta: CommitMeta = desert.ib(fields.Nested(CommitMetaSchema, data_key="commitMeta"))
    parent_commit_hash: str = attr.ib(default=None, metadata=desert.metadata(fields.Str(allow_none=True, data_key="parentCommitHash")))
    operations: List[Operation] = desert.ib(marshmallow_field=fields.List(fields.Nested(OperationsSchema()), allow_none=True), default=None)


LogEntrySchema = desert.schema_class(LogEntry)


@attr.dataclass
class LogResponse:
    """Dataclass for Log Response."""

    log_entries: List[LogEntry] = desert.ib(fields.List(fields.Nested(LogEntrySchema()), data_key="logEntries"))
    has_more: bool = attr.ib(default=False, metadata=desert.metadata(fields.Bool(allow_none=True, data_key="hasMore")))
    token: str = attr.ib(default=None, metadata=desert.metadata(fields.Str(allow_none=True)))


LogResponseSchema = desert.schema_class(LogResponse)


@attr.dataclass
class Transplant:
    """Dataclass for Transplant operation."""

    from_ref_name: str = attr.ib(metadata=desert.metadata(fields.Str(data_key="fromRefName")))
    hashes_to_transplant: List[str] = attr.ib(metadata=desert.metadata(fields.List(fields.Str(), data_key="hashesToTransplant")))
    keep_individual_commits: bool = attr.ib(
        default=True, metadata=desert.metadata(fields.Bool(allow_none=True, data_key="keepIndividualCommits"))
    )


TransplantSchema = desert.schema_class(Transplant)


@attr.dataclass
class Merge:
    """Dataclass for Merge operation."""

    from_ref_name: str = attr.ib(metadata=desert.metadata(fields.Str(data_key="fromRefName")))
    from_hash: str = attr.ib(default=None, metadata=desert.metadata(fields.Str(data_key="fromHash")))


MergeSchema = desert.schema_class(Merge)


@attr.dataclass
class MergeResponseDetails:
    """Dataclass for details in MergeResponse."""

    key: ContentKey = attr.ib(metadata=desert.metadata(fields.Nested(ContentKeySchema())))
    merge_behaviour: str = attr.ib(metadata=desert.metadata(fields.Str(data_key="mergeBehavior")))
    conflict_type: str = attr.ib(metadata=desert.metadata(fields.Str(data_key="conflictType")))
    source_commits: List[str] = attr.ib(metadata=desert.metadata(fields.List(fields.Str(), data_key="sourceCommits")))
    target_commits: List[str] = attr.ib(metadata=desert.metadata(fields.List(fields.Str(allow_none=True), data_key="targetCommits")))


MergeResponseDetailsSchema = desert.schema_class(MergeResponseDetails)


@attr.dataclass
class MergeResponse:
    """Dataclass for a MergeResponse."""

    target_branch: str = attr.ib(metadata=desert.metadata(fields.Str(data_key="targetBranch")))
    effective_target_hash: str = attr.ib(metadata=desert.metadata(fields.Str(data_key="effectiveTargetHash")))
    source_commits: List[LogEntry] = attr.ib(
        metadata=desert.metadata(fields.List(fields.Nested(LogEntrySchema()), data_key="sourceCommits"))
    )
    details: List[MergeResponseDetails] = attr.ib(
        metadata=desert.metadata(fields.List(fields.Nested(MergeResponseDetailsSchema()), data_key="details"))
    )
    resultant_target_hash: str = attr.ib(metadata=desert.metadata(fields.Str(allow_none=True, data_key="resultantTargetHash")))
    common_ancestor: str = attr.ib(metadata=desert.metadata(fields.Str(allow_none=True, data_key="commonAncestor")))
    expected_hash: str = attr.ib(metadata=desert.metadata(fields.Str(allow_none=True, data_key="expectedHash")))
    target_commits: List[LogEntry] = attr.ib(
        metadata=desert.metadata(fields.List(fields.Nested(LogEntrySchema()), allow_none=True, data_key="targetCommits"))
    )
    was_applied: bool = attr.ib(default=False, metadata=desert.metadata(fields.Bool(data_key="wasApplied")))
    was_successful: bool = attr.ib(default=False, metadata=desert.metadata(fields.Bool(data_key="wasSuccessful")))


MergeResponseSchema = desert.schema_class(MergeResponse)


@attr.dataclass
class MultiContents:
    """Contents container for commit."""

    commit_meta: CommitMeta = desert.ib(fields.Nested(CommitMetaSchema, data_key="commitMeta"))
    operations: List[Operation] = desert.ib(fields.List(fields.Nested(OperationsSchema())))


MultiContentSchema = desert.schema_class(MultiContents)


@attr.dataclass
class DiffEntry:
    """Dataclass for a Diff."""

    content_key: ContentKey = desert.ib(fields.Nested(ContentKeySchema, data_key="key"))
    from_content: Content = desert.ib(fields.Nested(ContentSchema, default=None, data_key="from", allow_none=True))
    to_content: Content = desert.ib(fields.Nested(ContentSchema, default=None, data_key="to", allow_none=True))

    def pretty_print(self) -> str:
        """Print out for cli."""
        # pylint: disable=E1101
        from_output = f"{self.from_content.pretty_print()}" if self.from_content else ""
        to_output = f"{self.to_content.pretty_print()}" if self.to_content else ""
        return (
            f"ContentKey: {self.content_key.to_path_string()}\nFROM:\n\t{from_output}\nTO:\n{to_output}"
            f"\n----------------------------------------"
        )


DiffEntrySchema = desert.schema_class(DiffEntry)


@attr.dataclass
class DiffResponse:
    """Dataclass for a DiffResponse."""

    diffs: List[DiffEntry] = desert.ib(fields.List(fields.Nested(DiffEntrySchema(), data_key="diffs")))


DiffResponseSchema = desert.schema_class(DiffResponse)


@attr.dataclass
class ReflogEntry:
    """Dataclass for reflog entries."""

    reflog_id: str = attr.ib(metadata=desert.metadata(fields.Str(data_key="refLogId")))
    ref_name: str = attr.ib(metadata=desert.metadata(fields.Str(data_key="refName")))
    ref_type: str = attr.ib(metadata=desert.metadata(fields.Str(data_key="refType")))
    commit_hash: str = attr.ib(metadata=desert.metadata(fields.Str(data_key="commitHash")))
    parent_reflog_id: str = attr.ib(metadata=desert.metadata(fields.Str(data_key="parentRefLogId")))
    operation_time: int = attr.ib(metadata=desert.metadata(fields.Int(data_key="operationTime")))
    operation: str = attr.ib(metadata=desert.metadata(fields.Str(data_key="operation")))
    source_hashes: List[str] = attr.ib(metadata=desert.metadata(fields.List(fields.Str(), data_key="sourceHashes")))


ReflogEntrySchema = desert.schema_class(ReflogEntry)


@attr.dataclass
class ReflogResponse:
    """Dataclass for reflog Response."""

    log_entries: List[ReflogEntry] = desert.ib(fields.List(fields.Nested(ReflogEntrySchema()), data_key="logEntries"))
    has_more: bool = attr.ib(default=False, metadata=desert.metadata(fields.Bool(allow_none=True, data_key="hasMore")))
    token: str = attr.ib(default=None, metadata=desert.metadata(fields.Str(allow_none=True)))


ReflogResponseSchema = desert.schema_class(ReflogResponse)



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Branch and tag common functions."""

from typing import Optional

import click

from pynessie.client import NessieClient
from pynessie.error import NessieConflictException
from pynessie.model import Branch, ReferenceSchema, Tag, split_into_reference_and_hash


def handle_branch_tag(
    nessie: NessieClient,
    list_references: bool,
    delete_reference: bool,
    ref_name: str,
    hash_on_ref: Optional[str],
    base_ref: str,
    is_branch: bool,
    json: bool,
    force: bool,
    verbose: bool,
    fetch_all: bool,
    expected_hash: Optional[str] = None,
) -> str:
    """Perform branch/tag actions.

    :param nessie NessieClient to use
    :param list_references the -l option choice
    :param delete_reference the -d option choice
    :param ref_name the name of the reference to create/assign/delete
    :param hash_on_ref a specific hash (reachable from base_ref) to be used for creating the new branch or tag instead
           of the HEAD of base_ref
    :param base_ref existing branch or tag to act at the base for creating the new branch or tag
    :param is_branch whether the operation is about a branch (true) or tag (false)
    :param json the --json option choice
    :param force the -f option choice
    :param verbose the -v option choice
    :param fetch_all the -x option to fetch additional metadata for a branch/tag
    :param expected_hash hash whose existence needs to be checked (on the server side) before performing the operation
    """
    if list_references or (not list_references and not delete_reference and not ref_name and not base_ref):
        return _handle_list(nessie, json, verbose, is_branch, ref_name, fetch_all)
    if delete_reference:
        if not hash_on_ref:
            hash_on_ref = nessie.get_reference(ref_name).hash_ or "fail"
        getattr(nessie, "delete_{}".format("branch" if is_branch else "tag"))(ref_name, hash_on_ref)
    else:  # create or force assign
        # use locally configured default branch as base_ref by default
        if not base_ref:
            base_ref = nessie.get_default_branch()
            hash_on_ref_arg = None
        else:
            base_ref, hash_on_ref_arg = split_into_reference_and_hash(base_ref)

        # use the current HEAD of base_ref as the hash for the new branch/tag by default
        if not hash_on_ref:
            hash_on_ref = hash_on_ref_arg if hash_on_ref_arg else nessie.get_reference(base_ref).hash_
        elif hash_on_ref_arg and hash_on_ref_arg != hash_on_ref:
            raise click.exceptions.BadOptionUsage(
                "hash_on_ref", "commit-hash provided via 'base-ref' and 'hash-on-ref', use only one of those"
            )

        # try creating a _new_ branch/tag first
        try:
            getattr(nessie, "create_{}".format("branch" if is_branch else "tag"))(ref_name, base_ref, hash_on_ref)
        except NessieConflictException as conflict:
            # NessieConflictException means the branch/tag already exists - force reassignment if requested
            if force:
                getattr(nessie, "assign_{}".format("branch" if is_branch else "tag"))(ref_name, base_ref, hash_on_ref, expected_hash)
            else:
                raise conflict
    return ""


def _handle_list(nessie: NessieClient, json: bool, verbose: bool, is_branch: bool, ref_name: str, fetch_all: bool) -> str:
    results = nessie.list_references(fetch_all=fetch_all).references
    kept_results = [ref for ref in results if isinstance(ref, (Branch if is_branch else Tag))]
    if ref_name:
        kept_results = [i for i in kept_results if i.name == ref_name]
    if json:
        return _handle_json_output(kept_results, ref_name)
    return _handle_normal_output(kept_results, verbose, nessie.get_default_branch(), fetch_all)


def _handle_json_output(input_data: list, ref_name: str) -> str:
    if ref_name and len(input_data) == 1:
        return ReferenceSchema().dumps(input_data[0], many=False)
    if ref_name and len(input_data) < 1:
        return "{}"
    return ReferenceSchema().dumps(input_data, many=True)


def _handle_normal_output(input_data: list, verbose: bool, default_branch: str, show_additional_info: bool = False) -> str:
    output = ""
    max_width = max((len(i.name) for i in input_data), default=0)
    for x in input_data:
        additional_info = ""
        if show_additional_info and x.metadata:
            commits_ahead = x.metadata.num_commits_ahead if x.metadata.num_commits_ahead else "?"
            commits_behind = x.metadata.num_commits_behind if x.metadata.num_commits_behind else "?"
            total_commits = x.metadata.num_total_commits if x.metadata.num_total_commits else "?"
            ancestor = x.metadata.common_ancestor_hash if x.metadata.common_ancestor_hash else "?"
            msg = (
                x.metadata.commit_meta_of_head.message[:35] + "..."
                if len(x.metadata.commit_meta_of_head.message) >= 35
                else x.metadata.commit_meta_of_head.message
            )
            head = (
                f"[author: {x.metadata.commit_meta_of_head.author} | commitTime: {x.metadata.commit_meta_of_head.commitTime} | msg: {msg}]"
                if x.metadata.commit_meta_of_head
                else "?"
            )
            additional_info = (
                f" [total commits: {total_commits} | commits behind: {commits_behind} | "
                f"commits ahead: {commits_ahead} | ancestor: {ancestor} | HEAD: {head}]"
            )

        next_row = "{}{}{}{}{}\n".format(
            "*".ljust(2) if x.name == default_branch else "  ",
            x.name.ljust(max_width + 1),
            " " if verbose else "",
            x.hash_ if verbose else "",
            additional_info,
        )
        output += click.style(next_row, fg="yellow") if x.name == default_branch else next_row
    return output



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Config based auth mechanism."""

from typing import Optional

from confuse import Configuration
from requests.auth import AuthBase, HTTPBasicAuth

from pynessie.auth.aws import setup_aws_auth
from pynessie.auth.bearer import TokenAuth as BearerTokenAuth


def setup_auth(config: Configuration) -> Optional[AuthBase]:
    """Setup an AuthBase object for future authentication against a Nessie Server.

    :param config: confuse.Configuration object
    :return: a requests.auth.AuthBase object that is able to inject authentication data into HTTP requests
    """
    auth_type = config["auth"]["type"].get()

    if auth_type == "none":
        return None
    if auth_type == "basic":
        username = config["auth"]["username"].get()
        password = config["auth"]["password"].get()
        return HTTPBasicAuth(username, password)
    if auth_type == "bearer":
        token = config["auth"]["token"].get()
        return BearerTokenAuth(token)
    if auth_type == "aws":
        region = config["auth"]["region"].get()
        profile = config["auth"]["profile"].get()
        return setup_aws_auth(region=region, profile=profile)

    raise NotImplementedError("Unsupported authentication type: " + auth_type)



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test Nessie models."""
from assertpy import assert_that

from pynessie.model import ContentKey


def test_content_key_to_string() -> None:
    """Test ContentKey.to_string method."""
    assert_that(ContentKey(["a", "b", "c"]).to_string()).is_equal_to("a.b.c")
    assert_that(ContentKey(["a", "b", "c.d"]).to_string()).is_equal_to('a.b."c.d"')
    assert_that(ContentKey(["a.b", "c", "d"]).to_string()).is_equal_to('"a.b".c.d')
    assert_that(ContentKey(["a.b", "c", "d.e"]).to_string()).is_equal_to('"a.b".c."d.e"')
    assert_that(ContentKey(["a", "b.c", "d"]).to_string()).is_equal_to('a."b.c".d')


def test_content_key_from_path_string() -> None:
    """Test ContentKey.from_path_string method."""
    assert_that(ContentKey.from_path_string("a.b.c")).is_equal_to(ContentKey(["a", "b", "c"]))
    assert_that(ContentKey.from_path_string("a.b.c\00d")).is_equal_to(ContentKey(["a", "b", "c.d"]))
    assert_that(ContentKey.from_path_string("a\00b.c.d")).is_equal_to(ContentKey(["a.b", "c", "d"]))
    assert_that(ContentKey.from_path_string("a\00b.c.d\00e")).is_equal_to(ContentKey(["a.b", "c", "d.e"]))
    assert_that(ContentKey.from_path_string("a.b\00c.d")).is_equal_to(ContentKey(["a", "b.c", "d"]))


def test_content_key_to_path_string() -> None:
    """Test ContentKey.to_path_string method."""
    assert_that(ContentKey(["a", "b", "c"]).to_path_string()).is_equal_to("a.b.c")
    assert_that(ContentKey(["a", "b", "c.d"]).to_path_string()).is_equal_to("a.b.c\00d")
    assert_that(ContentKey(["a.b", "c", "d"]).to_path_string()).is_equal_to("a\00b.c.d")
    assert_that(ContentKey(["a.b", "c", "d.e"]).to_path_string()).is_equal_to("a\00b.c.d\00e")
    assert_that(ContentKey(["a", "b.c", "d"]).to_path_string()).is_equal_to("a.b\00c.d")



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""cherry-pick CLI command."""

from typing import Tuple

import click
from click import UsageError

from pynessie.cli_common_context import ContextObject, MutuallyExclusiveOption
from pynessie.decorators import error_handler, pass_client, validate_reference


@click.command("cherry-pick")
@click.option("-b", "--branch", "ref", help="branch to cherry-pick onto. If not supplied the default branch from config is used")
@click.option(
    "-f",
    "--force",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    mutually_exclusive=["condition"],
    help="force branch assignment",
)
@click.option(
    "-c",
    "--condition",
    "expected_hash",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["force"],
    help="Expected hash. Only perform the action if the branch currently points to the hash specified by this option.",
)
@click.option("-s", "--source-ref", required=True, help="Name of the reference used to read the hashes from.")
@click.argument("hashes", nargs=-1, required=False)
@pass_client
@error_handler
@validate_reference
def cherry_pick(ctx: ContextObject, ref: str, force: bool, expected_hash: str, source_ref: str, hashes: Tuple[str]) -> None:
    """Cherry-pick HASHES onto another branch.

    HASHES commit hashes to be cherry-picked from the source reference.

    Examples:

        nessie cherry-pick -c 12345678abcdef -s dev 21345678abcdef 31245678abcdef -> cherry pick 2 commits with
    commit hash '21345678abcdef' '31245678abcdef' from dev branch to default branch
    with default branch's expected hash '12345678abcdef'

        nessie cherry-pick -b main -c 12345678abcdef -s dev 21345678abcdef 31245678abcdef -> cherry pick 2 commits with
    commit hash '21345678abcdef' '31245678abcdef' from dev branch to a branch named main
    with main branch's expected hash '12345678abcdef'
    """
    if not force and not expected_hash:
        raise UsageError(
            """Either condition or force must be set. Condition should be set to a valid hash for concurrency
            control or force to ignore current state of Nessie Store."""
        )
    ctx.nessie.cherry_pick(ref, source_ref, expected_hash, *hashes)
    click.echo()



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# flake8: noqa
"""The setup script."""
from setuptools import find_packages
from setuptools import setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

# here we have listed all dependencies w/o explicit pins to enable flexibility in client installs.
# we use `requirements.txt` in this directory when testing to ensure a stable test in CI.
requirements = [
    "attrs",  # features we use are not regularly changing
    "botocore",  # features we use are not regularly changing
    "Click<9.0.0,>6.0.0",  # pinning to 7.x or 8.x as we have used w/ both
    "confuse==2.0.1",  # important for config so don't change w/o testing
    "desert",  # features we use are not regularly changing
    "marshmallow",  # features we use are not regularly changing
    "marshmallow_oneofschema",  # features we use are not regularly changing
    "python-dateutil",  # stable
    "requests",  # stable
    "requests-aws4auth",  # stable
    "simplejson",  # stable
]

setup(
    author="Ryan Murray",
    author_email="nessie-release-builder@dremio.com",
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="Project Nessie: Transactional Catalog for Data Lakes with Git-like semantics",
    entry_points={
        "console_scripts": [
            "nessie=pynessie.cli:cli",
        ],
    },
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + "\n" + history,
    include_package_data=True,
    keywords="pynessie",
    name="pynessie",
    packages=find_packages(include=["pynessie", "pynessie.*"]),
    url="https://github.com/projectnessie/nessie",
    version="0.61.0",
    zip_safe=False,
)



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Remote CLI group command."""

import click

from pynessie.cli_common_context import ContextObject
from pynessie.conf import process
from pynessie.decorators import error_handler, pass_client


@click.group()
@pass_client
def remote(ctx: ContextObject) -> None:
    """Set and view remote endpoint."""
    pass


@remote.command("show")
@pass_client
@error_handler
def show(ctx: ContextObject) -> None:
    """Show current remote."""
    click.echo("Remote URL: " + ctx.nessie.get_base_url())
    click.echo("Default branch: " + ctx.nessie.get_reference(None).name)
    click.echo("Remote branches: ")
    for i in ctx.nessie.list_references().references:
        click.echo("\t" + i.name)


@remote.command(name="add")
@click.argument("endpoint", nargs=1, required=True)
@pass_client
@error_handler
def set_(ctx: ContextObject, endpoint: str) -> None:
    """Set current remote."""
    click.echo(process(None, "endpoint", False, None, endpoint))


@remote.command("set-head")
@click.argument("head", nargs=1, required=True)
@click.option("-d", "--delete", is_flag=True, help="delete the default branch")
@pass_client
@error_handler
def set_head(ctx: ContextObject, head: str, delete: bool) -> None:
    """Set current default branch. If -d is passed it will remove the default branch."""
    if delete:
        click.echo(process(None, None, False, "default_branch", None))
    else:
        click.echo(process(None, "default_branch", False, None, head))



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests for pynessie package setup."""

import os

from assertpy import assert_that


def test_all_valid_modules_to_have_init() -> None:
    """Test if all folder in pynessie have the required '__init__.py' file in order to treat them as modules."""
    pynessie_dir = os.path.join(os.getcwd(), "pynessie")
    _test_if_file_exists_nested_folders(pynessie_dir, "__init__.py")


def _test_if_file_exists_nested_folders(input_folder: str, expected_file_name: str) -> None:
    for file in os.listdir(input_folder):
        file_full_path = os.path.join(input_folder, file)
        if os.path.isdir(file_full_path) and file not in "__pycache__":
            assert_that(os.path.join(file_full_path, expected_file_name)).exists()
            _test_if_file_exists_nested_folders(file_full_path, expected_file_name)



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The purpose of this module is to provide different functions that can produce CEL Expressions based on given filtering parameters."""


from datetime import timezone
from typing import List, Optional

from dateutil import parser


def build_filter_for_commit_log_flags(
    query_filter: str, authors: List[str], committers: List[str], since: Optional[str], until: Optional[str]
) -> Optional[str]:
    """Producs a CEL expression to be used for filtering the commit log based on the given filtering parameters."""
    if query_filter:
        return query_filter
    expressions = []

    if authors:
        expressions.append(_expression_for_commit_log_by_author(authors))
    if committers:
        expressions.append(_expression_for_commit_log_by_committer(committers))
    if since:
        expressions.append(_expression_for_commit_log_by_commit_time_after(parse_to_iso8601(since)))
    if until:
        expressions.append(_expression_for_commit_log_by_commit_time_before(parse_to_iso8601(until)))

    if len(expressions) > 0:
        return _and_join(expressions)

    return None


def build_filter_for_contents_listing_flags(query_filter: str, types: List[str]) -> Optional[str]:
    """Producs a CEL expression to be used for filtering the content entries based on the given filtering parameters."""
    if query_filter:
        return query_filter

    if types:
        return _expression_for_entries_by_types(types)

    return None


def _expression_for_commit_log_by_author(authors: List[str]) -> str:
    return __generate_expression(authors, "commit.author=='{}'")


def _expression_for_commit_log_by_committer(committers: List[str]) -> str:
    return __generate_expression(committers, "commit.committer=='{}'")


def _expression_for_commit_log_by_commit_time_after(commit_time_after: str) -> str:
    return "timestamp(commit.commitTime) > timestamp('{}')".format(commit_time_after)


def _expression_for_commit_log_by_commit_time_before(commit_time_before: str) -> str:
    return "timestamp(commit.commitTime) < timestamp('{}')".format(commit_time_before)


def _expression_for_entries_by_types(types: List[str]) -> str:
    return "entry.contentType in [{}]".format(",".join(["'" + t + "'" for t in types]))


def _and_join(expressions: List[str]) -> str:
    result = " && ".join(expressions)
    if len(expressions) > 1:
        result = "(" + result + ")"
    return result


def _or_join(expressions: List[str]) -> str:
    result = " || ".join(expressions)
    if len(expressions) > 1:
        result = "(" + result + ")"
    return result


def __generate_expression(items: List[str], expression_template: str) -> str:
    expressions = []
    for item in items:
        expressions.append(expression_template.format(item))
    return _or_join(expressions)


def parse_to_iso8601(date_str: str) -> str:
    """Parses a given Date string to an ISO 8601 compliant string in UTC.

    :param date_str A string representation of a date
    :return An ISO 8601 string in the form YYYY-MM-DD HH:MM:SS.mmmmmm+HH:MM
    """
    try:
        return parser.parse(date_str).astimezone(tz=timezone.utc).isoformat(sep="T")
    except Exception as e:
        raise ValueError("Unable to parse '{}' to an ISO 8601-compliant string. Reason: {}".format(date_str, e), e) from e



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Branch CLI command."""

import click

from pynessie.cli_common_context import ContextObject, MutuallyExclusiveOption
from pynessie.commands._branch_tag_handlers import handle_branch_tag
from pynessie.decorators import error_handler, pass_client


@click.command(name="branch")
@click.option("-l", "--list", "is_list", cls=MutuallyExclusiveOption, is_flag=True, help="list branches", mutually_exclusive=["delete"])
@click.option("-d", "--delete", cls=MutuallyExclusiveOption, is_flag=True, help="delete a branch", mutually_exclusive=["is_list"])
@click.option("-f", "--force", is_flag=True, help="force branch assignment")
@click.option(
    "-o",
    "--hash-on-ref",
    help="Hash on source-reference for 'create' and 'assign' operations, if the branch shall not point to the HEAD of "
    "the given source-reference.",
)
@click.option(
    "-c",
    "--condition",
    "expected_hash",
    help="Expected hash. Only perform the action if the branch currently points to the hash specified by this option.",
)
@click.option(
    "-x",
    "--extended",
    "fetch_all",
    is_flag=True,
    help="Retrieve additional metadata for a branch, such as number of commits ahead/behind, "
    "info about the HEAD commit, number of total commits, or the common ancestor hash.",
)
@click.argument("branch", nargs=1, required=False)
@click.argument("base_ref", nargs=1, required=False)
@pass_client
@error_handler
def branch_(
    ctx: ContextObject,
    is_list: bool,
    force: bool,
    hash_on_ref: str,
    delete: bool,
    branch: str,
    base_ref: str,
    expected_hash: str,
    fetch_all: bool,
) -> None:
    """Branch operations.

    BRANCH name of branch to list or create/assign

    BASE_REF name of branch or tag from which to create/assign the new BRANCH

    Examples:

        nessie branch -> list all branches

        nessie branch -l -> list all branches

        nessie branch -l main -> list only main

        nessie branch -d main -> delete main

        nessie branch new_branch -> create new branch named 'new_branch' at current HEAD of the default branch

        nessie branch new_branch main -> create new branch named 'new_branch' at head of reference named 'main'

        nessie branch -o 12345678abcdef new_branch main -> create a branch named 'new_branch' at hash 12345678abcdef
    on reference named 'main'

        nessie branch new_branch main@12345678abcdef -> create a branch named 'new_branch' at hash 12345678abcdef
    on reference named 'main', alternate syntax for the above

        nessie branch -f existing_branch main -> assign branch named 'existing_branch' to head of reference named 'main'

        nessie branch -o 12345678abcdef -f existing_branch main -> assign branch named 'existing_branch' to hash 12345678abcdef
    on reference named 'main'

    """
    results = handle_branch_tag(
        ctx.nessie, is_list, delete, branch, hash_on_ref, base_ref, True, ctx.json, force, ctx.verbose, fetch_all, expected_hash
    )
    if ctx.json:
        click.echo(results)
    elif results:
        click.echo(results)
    else:
        click.echo()



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Error handling tests for Nessie CLI."""

import pytest

from .conftest import execute_cli_command

# Note: tests in this file use custom VCR files for error server responses.
# Running these tests in recording mode will likely NOT produce the expected
# server responses. The related VCR files need to be reviewed and corrected
# manually.


@pytest.mark.vcr(record_mode="none")
def test_server_error_html() -> None:
    """Test the handling of 500 responses with HTML payload (unexpected, but possible)."""
    result = execute_cli_command(["--json", "remote", "show"], ret_val=1)
    assert "Internal Server Error" in result
    assert "500" in result


@pytest.mark.vcr(record_mode="none")
def test_server_error_json() -> None:
    """Test the handling of 500 responses with JSON payload."""
    result = execute_cli_command(["--json", "remote", "show"], ret_val=1)
    assert "Default branch isn't a branch" in result
    assert "500" in result
